// ======================================================================
//
// Quaternion.cpp
// Gage Galinger
//
// Copyright 1998 Origin Systems Inc
//
// ======================================================================

#include "First.h"
#include "Quaternion.h"

#include "Xform.h"
#include "Vector.h"

const real QUATERNION_SLERP_EPSILON = CONST_REAL(0.1);

const Quaternion Quaternion::identity;

//****************************************************************************
//*****************  CONSTRUCTION  *******************************************
//****************************************************************************

Quaternion::Quaternion(void)
{
	loadIdentity();
}

Quaternion::Quaternion(const Quaternion& Q)
{
	setRotation(Q);
}

Quaternion::Quaternion(const Xform& X)
{
	setRotation(X);
}

Quaternion::Quaternion (real radians, const Vector& V)
{
	setRotation (radians, V);
}

Quaternion::Quaternion(real radians, real ax, real ay, real az)
{
	setRotation(radians, ax, ay, az);
}

Quaternion::~Quaternion(void)
{
}

//****************************************************************************
//*****************  UTILITY  ************************************************
//****************************************************************************

void Quaternion::loadIdentity()
{
	w = CONST_REAL (1);
	x = CONST_REAL (0);
	y = CONST_REAL (0);
	z = CONST_REAL (0);
}

void Quaternion::multiply(const Quaternion& Q)
{
	//-- This needs optimization.
	//-- The S is for "slow"
	real SW = w*Q.w - (x*Q.x + y*Q.y + z*Q.z);
	real SX = (y*Q.z - z*Q.y) + w*Q.x + Q.w*x;
	real SY = (z*Q.x - x*Q.z) + w*Q.y + Q.w*y;
	real SZ = (x*Q.y - y*Q.x) + w*Q.z + Q.w*z;
	w = SW;
	x = SX;
	y = SY;
	z = SZ;
}

//****************************************************************************
//*****************  ROTATION  ***********************************************
//****************************************************************************
void Quaternion::setRotation(const Quaternion& Quat)
{
	w = Quat.w;
	x = Quat.x;
	y = Quat.y;
	z = Quat.z;
}

void Quaternion::setRotation(const Xform& X)
{
#define MTX X.matrix
/*
	//-- This algorithm is from GDM, Feb. 1998
	real  tr, s, q[4];
	int    i, j, k;
	int    nxt[3] = {1, 2, 0};

	tr = MTX[0][0] + MTX[1][1] + MTX[2][2];

	if (tr > 0.0) 
	{
		s = sqrt(tr + CONST_REAL (1));
		w = s / CONST_REAL (2);
		s = CONST_REAL (0.5) / s;
		x = (MTX[2][1] - MTX[1][2]) * s;
		y = (MTX[0][2] - MTX[2][0]) * s;
		z = (MTX[1][0] - MTX[0][1]) * s;
		return;
	}

	i = 0;
	if (MTX[1][1] > MTX[0][0]) i = 1;
	if (MTX[2][2] > MTX[i][i]) i = 2;
	j = nxt[i];
	k = nxt[j];
	s = (real) sqrt((MTX[i][i] - (MTX[j][j] + MTX[k][k])) + 1.0);

	q[i] = s * CONST_REAL (0.5);

	if (s != CONST_REAL (0)) s = CONST_REAL (0.5) / s;

	q[3] = (MTX[k][j] - MTX[j][k]) * s;
	q[j] = (MTX[j][i] + MTX[i][j]) * s;
	q[k] = (MTX[k][i] + MTX[i][k]) * s;

	x = q[0];
	y = q[1];
	z = q[2];
	w = q[3];
*/
	//-- This algorithm is from GDM, Feb. 1998
	float  tr, s, q[4];
	int    i, j, k;
	int    nxt[3] = {1, 2, 0};

	tr = MTX[0][0] + MTX[1][1] + MTX[2][2];

	if (tr > 0.0) 
	{
		s = (float)sqrt(tr + 1.0f);
		w = s / 2.0f;
		s = 0.5f / s;
		x = (MTX[2][1] - MTX[1][2]) * s;
		y = (MTX[0][2] - MTX[2][0]) * s;
		z = (MTX[1][0] - MTX[0][1]) * s;
		return;
	}

	i = 0;
	if (MTX[1][1] > MTX[0][0]) i = 1;
	if (MTX[2][2] > MTX[i][i]) i = 2;
	j = nxt[i];
	k = nxt[j];
	s = (float) sqrt((MTX[i][i] - (MTX[j][j] + MTX[k][k])) + 1.0);

	q[i] = s * 0.5f;

	if (s != 0.0f) s = 0.5f / s;

	q[3] = (MTX[k][j] - MTX[j][k]) * s;
	q[j] = (MTX[j][i] + MTX[i][j]) * s;
	q[k] = (MTX[k][i] + MTX[i][k]) * s;

	x = q[0];
	y = q[1];
	z = q[2];
	w = q[3];

#undef MTX
	return;
}

void Quaternion::setRotation(real radians, const Vector& V)
{
	setRotation (radians, V.x, V.y, V.z);
}

void Quaternion::setRotation(real degrees, real ax, real ay, real az)
{
/*
	//-- Normalize the vector
	real mag = sqrt(ax*ax + ay*ay + az*az);
	if (mag == CONST_REAL (0))
	{
		Fatal ("mag == CONST_REAL (0)");
		loadIdentity ();
		return;
	}

	ax /= mag;
	ay /= mag;
	az /= mag;

	//-- Optimize this.  In particular, replace sin, cos with fast equivalents.
	w = cos(radians/CONST_REAL (2));
	x = sin(radians/CONST_REAL (2))*ax;
	y = sin(radians/CONST_REAL (2))*ay;
	z = sin(radians/CONST_REAL (2))*az;
*/
	//-- Optimize this.  In particular, replace sin, cos with fast equivalents.
	#define SIN(degrees) ((float)sin((degrees)*3.14f/180.0f))
	#define COS(degrees) ((float)cos((degrees)*3.14f/180.0f))

	//-- Normalize the vector
	float mag = (float)sqrt(ax*ax + ay*ay + az*az);
//	_ASSERT(mag);
	ax /= mag;
	ay /= mag;
	az /= mag;

	w = COS(degrees/2.0f);
	x = SIN(degrees/2.0f)*ax;
	y = SIN(degrees/2.0f)*ay;
	z = SIN(degrees/2.0f)*az;

	#undef SIN
	#undef COS

}

void Quaternion::addRotation(const Quaternion& Q)
{
	multiply(Q);
}

void Quaternion::addRotation(const Xform& X)
{
	Quaternion Q(X);
	multiply(Q);
}

void Quaternion::addRotation(real radians, const Vector& V)
{
	addRotation (radians, V.x, V.y, V.z);
}

void Quaternion::addRotation(real radians, real ax, real ay, real az)
{
	Quaternion Q(radians, ax, ay, az);
	multiply(Q);
}

void Quaternion::getXform(Xform* X) const
{
#define MTX X->matrix
	//-- Optimize this.
	ZeroMemory(MTX, sizeof(MTX));
/*
	MTX[0][0] = CONST_REAL (1) - 2*y*y - 2*z*z;
	MTX[0][1] = 2*x*y - 2*w*z;
	MTX[0][2] = 2*x*z + 2*w*y;

	MTX[1][0] = 2*x*y + 2*w*z;
	MTX[1][1] = CONST_REAL (1) - 2*x*x - 2*z*z;
	MTX[1][2] = 2*y*z - 2*w*x;

	MTX[2][0] = 2*x*z - 2*w*y;
	MTX[2][1] = 2*y*z + 2*w*x;
	MTX[2][2] = 1-2*x*x - 2*y*y;

	MTX[3][3] = CONST_REAL (1);
*/
	MTX[0][0] = 1.0f - 2*y*y - 2*z*z;
	MTX[0][1] = 2*x*y - 2*w*z;
	MTX[0][2] = 2*x*z + 2*w*y;

	MTX[1][0] = 2*x*y + 2*w*z;
	MTX[1][1] = 1.0f - 2*x*x - 2*z*z;
	MTX[1][2] = 2*y*z - 2*w*x;

	MTX[2][0] = 2*x*z - 2*w*y;
	MTX[2][1] = 2*y*z + 2*w*x;
	MTX[2][2] = 1-2*x*x - 2*y*y;

	MTX[3][3] = 1.0f;

#undef MTX
}

void Quaternion::getAxis(real* radians, real* ax, real* ay, real* az) const
{
#define SMALL    CONST_REAL (0.001)

	real cos_angle  = w;
	real angle      = (real)acos(cos_angle)*CONST_REAL (2);
	real sin_angle  = (real)sqrt(CONST_REAL (1) - cos_angle * cos_angle);

	if (fabs(sin_angle) < SMALL)
		sin_angle = 1;

	*radians  = angle;
	*ax       = x/sin_angle;
	*ay       = y/sin_angle;
	*az       = z/sin_angle;

	if (fabs(*radians) < SMALL) *radians = CONST_REAL (0);
	if (fabs(*ax) < SMALL)       *ax = CONST_REAL (0);
	if (fabs(*ay) < SMALL)       *ay = CONST_REAL (0);
	if (fabs(*az) < SMALL)       *az = CONST_REAL (0);

	#undef SMALL
}

// ----------------------------------------------------------------------
// makes "this" quaternion:
// slerp: spherical linear interpolation between 2 quaternions
//        by shortest distance across the unit sphere
//        (unit) quaternion a is the starting orientation
//        (unit) quaternion b is the ending orientation
//        0.0 <= alpha <= 1.0 is the interpolation amount
//
// notes:
//  - spin is the number of spins to use
//    (0 => shortest distance across sphere)
//    (1 => one extra rotation)
//    (-1 => "wrong way round" sphere)
//  - spin defaults to 0

void Quaternion::slerp (const Quaternion& a, const Quaternion& b, real alpha, real spin)
{
	real sin_t,cos_t,beta,theta,phi;
	real one = CONST_REAL(1);

	cos_t = (a.x * b.x) + (a.y * b.y) + (a.z * b.z) + (a.w * b.w);

	bool bflip = false;
	if (cos_t < 0)
	{
		cos_t = -cos_t;
		bflip = true;
	}

	if (one - cos_t < QUATERNION_SLERP_EPSILON)
		beta = one - alpha;
	else
	{
		theta = acos(cos_t);

		phi = theta + spin * Pi;

		sin_t = sin(theta);

		beta = sin (theta - alpha * phi);
		beta = beta / sin_t;

		alpha = sin (alpha * phi);
		alpha /= sin_t;
	}

	if (bflip)
		alpha = -alpha;

	x = beta * a.x + alpha * b.x;
	y = beta * a.y + alpha * b.y;
	z = beta * a.z + alpha * b.z;
	w = beta * a.w + alpha * b.w;
}

/*
// ======================================================================
//
// file: $/prophecy/src/engine/3d/quatern.cpp
// (c) origin systems, 1996
// author: hugh david
// description: quaternion class
// notes:
//  - deals only with unit quaternions
// to do:
//  - optimize
//  - Xform to Quaternion function
//
// ======================================================================

#include	"First.h"
#include	"Quaternion.h"

#include "MiscMath.h"
#include "Vector.h"
#include "Xform.h"

// ======================================================================

Quaternion::Quaternion(const Xform &xform)
{
#if 1
	UNREF(xform);
	w = CONST_REAL(1);
	x = CONST_REAL(9);
	y = CONST_REAL(9);
	z = CONST_REAL(9);
#else
	// aa4 stands for 4.0 * q.a * q.a, where q is the quaternion we are looking for; similarly for bb4, cc4, and dd4.
	// At this point, 0 <= aa4 <= 1, or the matrix was invalid.
	// We are using the largest of aa4, bb4, cc4, dd4 to guarantee numeric stability.
	real aa4 =   xform.matrix[0][0] - xform.matrix[1][1] - xform.matrix[2][2] + xform.matrix[3][3];
	real bb4 = - xform.matrix[0][0] + xform.matrix[1][1] - xform.matrix[2][2] + xform.matrix[3][3];
	real cc4 = - xform.matrix[0][0] - xform.matrix[1][1] + xform.matrix[2][2] + xform.matrix[3][3];
	real dd4 =   xform.matrix[0][0] + xform.matrix[1][1] + xform.matrix[2][2] + xform.matrix[3][3];
	real maximum = max (aa4, max (bb4, max (cc4, dd4)));

	if (aa4 == maximum)
	{
		real a  = sqrt (aa4 / CONST_REAL(4));
		real ab = (xform.matrix[1][0] + xform.matrix[0][1]) / CONST_REAL(4);
		real ac = (xform.matrix[0][2] + xform.matrix[2][0]) / CONST_REAL(4);
		real ad = (xform.matrix[2][1] - xform.matrix[1][2]) / CONST_REAL(4);
		x = a;
		y = ab / a;
		z = ac / a;
		w = ad / a;
	}
	else
	if (bb4 == max)
	{
		b  = sqrt (bb4 / CONST_REAL(4));
		ab = (xform.matrix[1][0] + xform.matrix[0][1]) / CONST_REAL(4);
		bc = (xform.matrix[2][1] + xform.matrix[1][2]) / CONST_REAL(4);
		bd = (xform.matrix[0][2] - xform.matrix[2][0]) / CONST_REAL(4);
		x = ab / b;
		y = b;
		z = bc / b;
		w = bd / b;
	}
	else
	if (cc4 == max)
	{
		c  = sqrt (cc4 / CONST_REAL(4));
		ac = (xform.matrix[0][2] + xform.matrix[2][0]) / CONST_REAL(4);
		bc = (xform.matrix[2][1] + xform.matrix[1][2]) / CONST_REAL(4);
		cd = (xform.matrix[1][0] - xform.matrix[0][1]) / CONST_REAL(4);
		x = ac / c;
		y = bc / c;
		z = c;
		w = cd / c;
	}
	else
	if (dd4 == max)
	{
		d  = sqrt (dd4 / CONST_REAL(4));
		ad = (xform.matrix[2][1] - xform.matrix[1][2]) / CONST_REAL(4);
		bd = (xform.matrix[0][2] - xform.matrix[2][0]) / CONST_REAL(4);
		cd = (xform.matrix[1][0] - xform.matrix[0][1]) / CONST_REAL(4);
		x = ad / d;
		y = bd / d;
		z = cd / d;
		w = d;
	}
	else
		Fatal("Quaternion::Quaternion (const Xform& xform).\n");

	if (w < CONST_REAL(0))
	{
		x = -x;
		y = -y;
		z = -z;
		w = -w;
	}
#endif
}

// ----------------------------------------------------------------------
// this routine makes "this" quaternion suitable to rotate points or vectors
// around the given vector by the given angle:

void Quaternion::rotation (const Vector &v, real angle)
{
	real l = v.magnitude();

	if (l == 0)
	{
		// if Vector has no magnitude, the rotation operation is undefined.
		// the best thing to do is just make "this" Quaternion the identity:
		w = CONST_REAL(1);
		x = CONST_REAL(0);
		y = CONST_REAL(0);
		z = CONST_REAL(0);
		return;
	}

	real halfangle = angle / CONST_REAL(2);

	real s = sin(halfangle) / l;
	real c = cos(halfangle);

	x = v.x * s;
	y = v.y * s;
	z = v.z * s;
	w = c;
}

// ======================================================================
*/